from ftmq.io import smart_read_proxies, smart_write_proxies
from ftmq.query import Query
from ftmq.util import make_entity

__version__ = "4.5.5"
__all__ = [
    "smart_read_proxies",
    "smart_write_proxies",
    "Query",
    "make_entity",
]
